<?php

namespace NoccyLabs\Plugin\Extension;

use NoccyLabs\Plugin\PluginManager;
use NoccyLabs\Plugin\PluginInstance;

// --8<-- ConfigurationExtension.php

/**
 * Enables configuration
 *
 * $config = $plugin->getConfiguration()
 * if (!$config->has('hello')) { $config->set('hello','world'); }'
 *
 */
class ConfigurationExtension extends AbstractExtension
{

    public function getDefaultName()
    {
        return "config";
    }

    public function onRegister()
    {
        $this->getDataPath(""); // this will throw an exception and stop load if config not available
        $this->plugin->extend("getConfiguration", [ $this, "getConfiguration" ]);
    }

    public function getConfiguration()
    {
        $path = $this->getDataPath("config.json");
        $props = new Configuration($path);
        return $props;
    }

}

// --8<-- ConfigurationManager.php

class Configuration
{
    protected $path;

    protected $props = [];
    
    protected $propsRead = [];

    protected $dirty = false;

    public function __construct($path)
    {
        $this->path = $path;
        $this->readConfig();
    }

    public function __destruct()
    {
        if ($this->dirty) {
            $this->writeConfig();
        }
    }

    public function readConfig()
    {
        $this->dirty = false;
        if (!file_exists($this->path)) {
            return;
        }
        $json = file_get_contents($this->path);
        $this->propsRead = (array)json_decode($json);
    }

    public function writeConfig()
    {
        $dir = dirname($this->path);
        if (!is_dir($dir)) {
            mkdir($dir, 0777, true);
        }
        $props = [];
        foreach ($this->props as $prop) {
            $props[$prop->name] = $prop->value;
        }
        $json = json_encode($props, JSON_PRETTY_PRINT);
        file_put_contents($this->path, $json."\n");
        $this->dirty = false;
    }

    protected function createProp($type, $name, $default)
    {
        return (object)[
            'name' => $name,
            'type' => $type,
            'value' => (array_key_exists($name,$this->propsRead))?$this->propsRead[$name]:$default,
            'default' => $default
        ];
    }

    public function has($key)
    {
        return array_key_exists($key, $this->props);
    }

    public function get($key)
    {
        return $this->props[$key]->value;
    }

    public function set($key, $value)
    {
        $this->props[$key]->value = $value;
    }

    public function reset($key)
    {
        $this->props[$key]->value = $this->props[$key]->default;
    }

    public function addIntProp($name, $default=null)
    {
        if ($this->has($name)) return $this;
        $this->dirty = true;
        $this->props[$name] = $this->createProp("int",$name,$default);
        return $this;
    }

    public function addStringProp($name, $default=null)
    {
        if ($this->has($name)) return $this;
        $this->dirty = true;
        $this->props[$name] = $this->createProp("string",$name,$default);
        return $this;
    }

    public function addBoolProp($name, $default=null)
    {
        if ($this->has($name)) return $this;
        $this->dirty = true;
        $this->props[$name] = $this->createProp("bool",$name,$default);
        return $this;
    }

    public function addMixedProp($name, $default=null)
    {
        if ($this->has($name)) return $this;
        $this->dirty = true;
        $this->props[$name] = $this->createProp("mixed",$name,$default);
        return $this;
    }
}

