<?php

namespace NoccyLabs\Plugin;

class PluginManager
{

    protected $dataPath = null;

    protected $cachePath = null;

    protected $paths = [];

    protected $extensions;

    protected $plugins = [];

    protected $loaded = [];

    public function __construct()
    {
        $this->extensions = new Extension\ExtensionManager($this);
    }

    public function __destruct()
    {
        $this->unloadAll();
    }

    public function addExtension(Extension\AbstractExtension $extension, $name=null)
    {
        $this->extensions->addExtension($extension, $name);
    }

    public function unloadAll()
    {
        foreach ($this->loaded as $plugin=>$version) {
            $plg = $this->findPlugin($plugin, $version);
            if ($plg) {
                $plg->unload();
            }
        }
        $this->loaded = [];
    }

    public function getExtensionManager()
    {
        return $this->extensions;
    }

    /**
     *
     *
     * @param
     */
    public function setCacheDirectory($path)
    {
        $this->cachePath = $path;
    }

    public function getCacheDirectory()
    {
        return $this->cachePath;
    }

    public function setDataDirectory($path)
    {
        $this->dataPath = $path;
    }

    public function getDataDirectory()
    {
        return $this->dataPath;
    }

    /**
     *
     *
     * @param
     */
    public function addPluginDirectory($path)
    {
        if (!in_array($path, $this->paths)) {
            $this->paths[] = $path;
        }
    }

    /**
     *
     *
     * @param string $plugin
     * @return bool
     */
    public function isLoaded($plugin)
    {
        if ($plugin instanceof PluginInstance) {
            $plugin = $plugin->getName();
        }
        return array_key_exists($plugin, $this->loaded)?$this->loaded[$plugin]:false;
    }

    /**
     *
     *
     * @param string $plugin
     * @param string $version
     */
    public function setLoaded($plugin, $version)
    {
        $this->loaded[$plugin] = $version;
    }

    /**
     * Reload the plugin list. This will currently reset all plugin instances, so
     * you will have to call load() on the ones you need. Be aware that this can
     * lead to leaks if objects have been registered by the plugin.
     *
     */
    public function reload()
    {
        $this->unloadAll();

        $plugins = [];
        foreach ($this->paths as $path) {
            $plugins = array_merge($plugins, $this->reloadPath($path));
        }
        foreach ($plugins as $plugin) {
            $plugin->setManager($this);
        }
        $this->plugins = $plugins;
        $this->resolveDependencies();
    }

    /**
     * Check the dependencies for the plugins, making sure that required libraries
     * are present.
     *
     */
    protected function resolveDependencies()
    {
        // TODO: Pin dependencies such that the latest compatible version is
        // known in advance. This way two plugins using overlapping dependencies
        // can still get the latest supported by each. This also removes the need
        // for getPlugin() to attempt to resolve what version to return.
    
        foreach ($this->plugins as $plugin) {
            $requires = $plugin->getDependencies();
            $missing = [];
            $deps = [];
            foreach ($requires as $name=>$version) {
                $dependency = $this->findPlugin($name, $version);
                if (!$dependency) {
                    $missing[] = sprintf("%s (%s)", $name, $version);
                } else {
                    $deps[] = $dependency;
                }
            }
            if (count($missing)>0) {
                $plugin->setBroken("Missing dependencies: ".join(", ",$missing));
            } else {
                foreach ($deps as $dep) {
                    $dep->addDependent($plugin);
                }
            }
        }
    }

    /**
     *
     *
     */
    public function findPlugin($name, $version=null)
    {
        $match = null;
        if (strpos($version," ")!==false) {
            $versions = explode(" ",$version);
        } else {
            $versions = [ $version ];
        }
        $comps = [];
        foreach ($versions as $version) {
            if ($version==null || $version=="*") {
                $comps[] = [ "0.0.0", ">=" ];
            } else {
                for ($n=0;$n<strlen($version);$n++) {
                    if (ctype_digit($version[$n])) {
                        $vcomp = substr($version,$n);
                        $vop = substr($version,0,$n)?:"==";
                        switch ($vop) {
                            case '~':
                                $vcompnext = explode(".",$vcomp);
                                array_pop($vcompnext);
                                $vcompnext[0]++;
                                $vcompnext = join(".",$vcompnext);
                                $comps[] = [ $vcomp, ">=" ];
                                $comps[] = [ $vcompnext, "<" ];
                                break;
                            default:
                                $comps[] = [ $vcomp, $vop ];
                        }
                        break;
                    }
                }
            }
        }

        $matches = [];
        foreach ($this->plugins as $plugin) {
            if ($plugin->getName() != $name) {
                continue;
            }
            foreach ($comps as $comp) {
                if (!version_compare($plugin->getVersion(), $comp[0], $comp[1])) {
                    continue(2);
                }
            }
            $matches[$plugin->getVersion()] = $plugin;
            /*
            if ($match && version_compare($match->getVersion(), $plugin->getVersion(), "<")) {
                continue;
            }
            $match = $plugin;
            */
        }

        uksort($matches, "version_compare");

        return end($matches);
    }

    /**
     *
     *
     */
    public function reloadPath($path)
    {
        if (!is_dir($path)) {
            return [];
        }

        $plugins = [];
        $it = new \DirectoryIterator($path);
        foreach ($it as $item) {
            $path = $item->getPathname();
            if (basename($path)=="..") continue;
            if (basename($path)==".") $path = dirname($path);
            $loader = Loader\LoaderFactory::createLoader($path);
            if ($loader) {
                $plugins[] = new PluginInstance($loader);
            }
        }
        return $plugins;
    }

    /**
     * Evaluate a function for each of the known plugins.
     *
     */
    public function each(callable $func)
    {
        foreach ($this->plugins as $plugin) {
            call_user_func($func, $plugin, $this);
        }
    }

    /**
     *
     *
     */
    public function getPlugin($name)
    {
        // Return the loaded version if loaded, latest version otherwise
        if (array_key_exists($name, $this->loaded)) {
            $version = $this->loaded[$name];
        } else {
            $version = "*";
        }
        // Find the matching plugin
        return $this->findPlugin($name, $version);
    }

    public function getPluginNames()
    {
        $names = array_map($this->plugins, function ($plugin) {
            return $plugin->getName();
        });
        return array_unique($names);
    }

    public function getLoadedVersions()
    {
        return $this->loaded;
    }

    public function loadPlugin($name, $version=null)
    {
        $plugin = $this->findPlugin($name, $version);
        if (!$plugin) {
            throw new PluginManagerException("No such plugin {$name}".($version?":{$version}":""));
        }

        try {
            $plugin->load();
        } catch (PluginManagerException $e) {
            fprintf(STDERR, "Warning: %s\n", $e->getMessage());
            return false;
        }
        return true;
    }

}

